from typing import Any, Dict, List

from .errors import ExpectedParsingError, UnexpectedParsingError
from .scraper import Scraper
from .table_parser import TableParser
from .utils import get_day_month, parse_select, parse_table_fields_args


class Race(Scraper):
    """
    Scraper for race overview HTML page.

    Usage:

    >>> from procyclingstats import Race
    >>> race = Race("race/tour-de-france/2022")
    >>> race.enddate()
    '2022-07-24'
    >>> race.parse()
    {
        'category': 'Men Elite',
        'edition': 109,
        'enddate': '2022-07-24',
        'is_one_day_race': False,
        ...
    }

    """
    def year(self) -> int:
        """
        Parse year when the race occured from HTML.

        :return: Year when the race occured.
        """
        return int(self.html.css_first("span.hideIfMobile").text())

    def name(self) -> str:
        """
        Parses display name from HTML.

        :return: Name of the race, e.g. ``Tour de France``.
        """
        display_name_html = self.html.css_first(".page-title > .main > h1")
        return display_name_html.text()

    def is_one_day_race(self) -> bool:
        """
        Parses whether race is one day race from HTML.

        :return: Whether given race is one day race.
        """
        titles = self.html.css("div > div > h3")
        titles = [] if not titles else titles
        for title_html in titles:
            if "Stages" in title_html.text():
                return False
        return True

    def nationality(self) -> str:
        """
        Parses race nationality from HTML.

        :return: 2 chars long country code in uppercase.
        """
        nationality_html = self.html.css_first(
            ".page-title > .main > span.flag")
        flag_class = nationality_html.attributes['class']
        return flag_class.split(" ")[1].upper() # type: ignore

    def edition(self) -> int:
        """
        Parses race edition year from HTML.

        :return: Edition as int.
        """
        edition_html = self.html.css_first(
            ".page-title > .main > span + font")
        if edition_html is not None:
            return int(edition_html.text()[:-2])
        raise ExpectedParsingError("Race cancelled, edition unavailable.")

    def startdate(self) -> str:
        """
        Parses race startdate from HTML.

        :return: Startdate in ``DD-MM-YYYY`` format.
        """
        startdate_html = self.html.css_first(
            ".infolist > li > div:nth-child(2)")
        return startdate_html.text()

    def enddate(self) -> str:
        """
        Parses race enddate from HTML.

        :return: Enddate in ``DD-MM-YYYY`` format.
        """
        enddate_html = self.html.css(".infolist > li > div:nth-child(2)")[1]
        return enddate_html.text()

    def category(self) -> str:
        """
        Parses race category from HTML.

        :return: Race category e.g. ``Men Elite``.
        """
        category_html = self.html.css(".infolist > li > div:nth-child(2)")[2]
        return category_html.text()

    def uci_tour(self) -> str:
        """
        Parses UCI Tour of the race from HTML.

        :return: UCI Tour of the race e.g. ``UCI Worldtour``.
        """
        uci_tour_html = self.html.css(".infolist > li > div:nth-child(2)")[3]
        return uci_tour_html.text()

    def prev_editions_select(self) -> List[Dict[str, str]]:
        """
        Parses previous race editions from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        editions_select_html = self.html.css_first("form > select")
        return parse_select(editions_select_html)

    def stages(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses race stages from HTML (available only on stage races). When
        race is one day race, empty list is returned.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_url: Winner's URL.
            - rider_name: Winner's name.
            - nationality: Winner's nationality as 2 chars long country code.
            - date: Date when the stage occured in ``MM-DD`` format.
            - profile_icon: Profile icon of the stage (p1, p2, ... p5).
            - stage_name: Name of the stage, e.g \
                ``Stage 2 | Roskilde - Nyborg``.
            - stage_url: URL of the stage, e.g. \
                ``race/tour-de-france/2022/stage-2``.

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "date",
            "profile_icon",
            "stage_name",
            "stage_url",
            "rider_url",
            "rider_name",
            "nationality"
        )
        if self.is_one_day_race():
            return []

        fields = parse_table_fields_args(args, available_fields)
        stages_table_html = self.html.css_first("div:not(.mg_r2) > div > \
            span > table.basic")
        if not stages_table_html:
            return []
        # remove rest day table rows
        for stage_e in stages_table_html.css("tr"):
            not_p_icon = stage_e.css_first(".icon.profile.p")
            if not_p_icon:
                stage_e.remove()

        table_parser = TableParser(stages_table_html)
        casual_f_to_parse = [f for f in fields if f != "date"]
        try:
            table_parser.parse(casual_f_to_parse)
        # if nationalities don't fit stages winners
        except UnexpectedParsingError:
            casual_f_to_parse.remove("nationality")
            table_parser.parse(casual_f_to_parse)
            nats = table_parser.nationality()
            j = 0
            for i in range(len(table_parser.table)):
                if j < len(nats) and \
                    table_parser.table[i]['rider_url'].split("/")[1]:
                    table_parser.table[i]['nationality'] = nats[j]
                    j += 1
                else:
                    table_parser.table[i]['nationality'] = None

        # add stages dates to table if neede
        if "date" in fields:
            dates = table_parser.parse_extra_column(0, get_day_month)
            table_parser.extend_table("date", dates)
        return table_parser.table



from typing import Any, Dict, List

from .scraper import Scraper
from .table_parser import TableParser
from .utils import parse_table_fields_args


class RaceStartlist(Scraper):
    """
    Scraper for race startlist HTML page.

    Usage:

    >>> from procyclingstats import RaceStartlist
    >>> race_startlist = RaceStartlist("race/tour-de-france/2022/startlist")
    >>> race_startlist.startlist()
    [
        {
            'nationality': 'SI',
            'rider_name': 'POGAČAR Tadej',
            'rider_number': 1,
            'rider_url': 'rider/tadej-pogacar',
            'team_name': 'UAE Team Emirates',
            'team_url': 'team/uae-team-emirates-2022'}
        },
        ...
    ]
    >>> race_startlist.parse()
    {
        'startlist': [
            {
                'nationality': 'SI',
                'rider_name': 'POGAČAR Tadej',
                'rider_number': 1,
                'rider_url': 'rider/tadej-pogacar',
                'team_name': 'UAE Team Emirates',
                'team_url': 'team/uae-team-emirates-2022'}
            },
            ...
        ]
    }
    """
    def startlist(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses startlist from HTML. When startlist is individual (without
        teams) fields team name, team url and rider nationality are set to
        None.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name:
            - rider_url:
            - team_name:
            - team_url:
            - nationality: Rider's nationality as 2 chars long country code.
            - rider_number: Rider's ID number in the race. For races without
                numbered participants (e.g. the ones that haven't occured yet)
                is every rider's ID None.

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rider_name",
            "rider_url",
            "team_name",
            "team_url",
            "nationality",
            "rider_number"
        )
        fields = parse_table_fields_args(args, available_fields)
        startlist_html = self.html.css_first("table.basic")

        # if startlist is a table
        if startlist_html:
            startlist_parser = TableParser(startlist_html)
            casual_fields = [f for f in fields if f != "rider_number"]
            startlist_parser.parse(casual_fields)
            # adds rider number to table if needed
            if "rider_number" in fields:
                numbers = startlist_parser.parse_extra_column(0,
                    lambda x: int(x) if x else None)
                startlist_parser.extend_table("rider_number", numbers)
            return startlist_parser.table

        casual_rider_fields = [
            "rider_name",
            "rider_url",
            "nationality"
        ]
        table = []
        startlist_html = self.html.css_first(".startlist_v4")
        for team_html in startlist_html.css(".ridersCont"):
            riders_table = team_html.css_first("ul")
            table_parser = TableParser(riders_table)
            rider_f_to_parse = [f for f in casual_rider_fields if f in fields]
            table_parser.parse(rider_f_to_parse)
            # add rider numbers to the table if needed
            if "rider_number" in fields:
                numbers = []
                for row in riders_table.css("li > .bib"):
                    num = row.text(deep=False).split(" ")[0]
                    if num.isnumeric():
                        numbers.append(int(num))
                    else:
                        numbers.append(None)
                table_parser.extend_table("rider_number", numbers)
            # add team names to the table if needed
            if "team_name" in fields:
                team_name = team_html.css_first("a").text()
                team_names = [team_name for _ in range(
                    len(table_parser.table))]
                table_parser.extend_table("team_name", team_names)
            # add team urls to the table if needed
            if "team_url" in fields:
                team_url = team_html.css_first("a").attributes['href']
                team_urls = [team_url for _ in range(len(table_parser.table))]
                table_parser.extend_table("team_url", team_urls)
            # add team table to startlist table
            table.extend(table_parser.table)
        return table



import re
from typing import Any, Dict, List, Literal, Tuple

from .errors import ExpectedParsingError
from .scraper import Scraper
from .table_parser import TableParser
from .utils import parse_select, parse_table_fields_args, select_menu_by_name


class Ranking(Scraper):
    """
    Scraper for rankings HTML page.

    Always only one parsing method that parses ranking is availabe, the others
    raise `ExpectedParsingError`. E.g. for object created with example URL
    would be valid only `self.individual_ranking` parsing method and others
    methods that parse ranking (`self.team_ranking`, ...) would raise error.

    Usage:

    >>> from procyclingstats import Ranking
    >>> ranking = Ranking("rankings/me/individual")
    >>> ranking.individual_ranking()
    [
        {
            'nationality': 'SI',
            'points': 2981.0,
            'prev_rank': 1,
            'rank': 1,
            'rider_name': 'Pogačar Tadej',
            'rider_url': 'rider/tadej-pogacar',
            'team_name': 'UAE Team Emirates',
            'team_url': 'team/uae-team-emirates-2022'
        },
        ...
    ]
    >>> ranking.parse()
    {
        'dates_select': [
            {'text': '2022-09-11', 'value': '2022-09-11'},
            {'text': '2021-12-31', 'value': '2021-12-31'},
            {'text': '2019-12-31', 'value': '2019-12-31'},
            ...
        ],
        'distance_ranking': None,
        'individual_ranking': [
            {
                'nationality': 'SI',
                'points': 2981.0,
                'prev_rank': 1,
                'rank': 1,
                'rider_name': 'Pogačar Tadej',
                'rider_url': 'rider/tadej-pogacar',
                'team_name': 'UAE Team Emirates',
                'team_url': 'team/uae-team-emirates-2022'
            },
            ...
        ],
        'individual_wins_ranking': None,
        ...
    }
    """
    def individual_ranking(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses individual ranking from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name:
            - rider_url:
            - team_name:
            - team_url:
            - rank: Rider's rank in the ranking.
            - prev_rank: Rider's rank in previous ranking update.
            - nationality: Rider's nationality as 2 chars long country code.
            - points:

        :raises ExpectedParsingError: When the table from HTML is not an
            individual points ranking table.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rank",
            "prev_rank",
            "rider_name",
            "rider_url",
            "team_name",
            "team_url",
            "nationality",
            "points"
        )
        if self._ranking_type() != "individual":
            raise ExpectedParsingError(
                "This object doesn't support individual_ranking method, create"
                "one with individual ranking URL to call this method.")
        return self._parse_regular_ranking_table(args, available_fields)

    def team_ranking(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses team ranking from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - team_name:
            - team_url:
            - rank: Team's rank in the ranking.
            - prev_rank: Team's rank in previous ranking update.
            - nationality: Team's nationality as 2 chars long country code.
            - class: Team's class, e.g. ``WT``.
            - points:

        :raises ExpectedParsingError: When the table from HTML is not a team
            points ranking table.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rank",
            "prev_rank",
            "team_name",
            "team_url",
            "nationality",
            "class",
            "points"
        )
        if self._ranking_type() != "teams":
            raise ExpectedParsingError(
                "This object doesn't support team_ranking method, "
                "create one with teams ranking URL to call this method.")
        return self._parse_regular_ranking_table(args, available_fields)

    def nations_ranking(self, *args: str) -> List[Dict[str, Any]]:

        """
        Parses nations ranking from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - nation_name:
            - nation_url:
            - rank: Nation's rank in the ranking.
            - prev_rank: Nation's rank in previous ranking update.
            - nationality: Nation as 2 chars long country code.
            - points:

        :raises ExpectedParsingError: When the table from HTML is not a
            nationality points ranking table.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rank",
            "prev_rank",
            "nation_name",
            "nation_url",
            "nationality",
            "points"
        )
        if self._ranking_type() != "nations":
            raise ExpectedParsingError(
                "This object doesn't support nations_ranking method, create" +
                "one with nations ranking URL to call this method.")
        return self._parse_regular_ranking_table(args, available_fields)

    def races_ranking(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses race ranking from HTML. Race points are evaluated based on
            startlist quality score.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - race_name:
            - race_url:
            - rank: Race's rank in the ranking.
            - prev_rank: Race's rank in previous ranking update.
            - nationality: Race's nationality as 2 chars long country code.
            - class: Race's class, e.g. ``WT``.
            - points:

        :raises ExpectedParsingError: When the table from HTML is not a race
            ranking table.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rank",
            "prev_rank",
            "race_name",
            "race_url",
            "nationality",
            "class",
            "points"
        )
        if self._ranking_type() != "races":
            raise ExpectedParsingError(
                "This object doesn't support races_ranking method, create one"
                "with race ranking URL to call this method.")

        fields = parse_table_fields_args(args, available_fields)
        html_table = self.html.css_first("table")
        table_parser = TableParser(html_table)
        # parse race name and url as stage name and url and rename it
        # afterwards
        if "race_name" in fields:
            fields[fields.index("race_name")] = "stage_name"
        if "race_url" in fields:
            fields[fields.index("race_url")] = "stage_url"
        table_parser.parse(fields)
        table_parser.rename_field("stage_name", "race_name")
        table_parser.rename_field("stage_url", "race_url")
        return table_parser.table

    def individual_wins_ranking(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses individual wins ranking from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name:
            - rider_url:
            - team_name:
            - team_url:
            - rank: Rider's rank in the ranking.
            - prev_rank: Rider's rank in previous ranking update.
            - nationality: Rider's nationality as 2 chars long country code.
            - first_places:
            - second_places:
            - third_places:

        :raises ExpectedParsingError: When the table from HTML is not an
            individual wins ranking table.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rank",
            "prev_rank",
            "rider_name",
            "rider_url",
            "team_name",
            "team_url",
            "nationality",
            "first_places",
            "second_places",
            "third_places"
        )
        if self._ranking_type() != "individual_wins":
            raise ExpectedParsingError(
                "This object doesn't support races_ranking method, create one"
                "with individual wins ranking URL to call this method.")
        return self._parse_regular_ranking_table(args, available_fields)

    def teams_wins_ranking(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses team wins ranking from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - team_name:
            - team_url:
            - rank: Team's rank in the ranking.
            - prev_rank: Team's rank in previous ranking update.
            - nationality: Team's nationality as 2 chars long country code.
            - class: Team's class, e.g. ``WT``.
            - first_places:
            - second_places:
            - third_places:

        :raises ExpectedParsingError: When the table from HTML is not a team
            wins ranking.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rank",
            "prev_rank",
            "team_name",
            "team_url",
            "nationality",
            "class",
            "first_places",
            "second_places",
            "third_places"
        )
        if self._ranking_type() != "team_wins":
            raise ExpectedParsingError(
                "This object doesn't support teams_wins_ranking method, "
                "create one with teams wins ranking URL to call this method.")
        return self._parse_regular_ranking_table(args, available_fields)

    def nations_wins_ranking(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses nations wins ranking from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - nation_name:
            - nation_url:
            - rank: Nation's rank in the ranking.
            - prev_rank: Nation's rank in previous ranking update.
            - nationality: Nation as 2 chars long country code.
            - first_places:
            - second_places:
            - third_places:

        :raises ExpectedParsingError: When the table from HTML is not a nation
            wins ranking table.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rank",
            "prev_rank",
            "nation_name",
            "nation_url",
            "nationality",
            "first_places",
            "second_places",
            "third_places"
        )

        if self._ranking_type() != "nation_wins":
            raise ExpectedParsingError(
                "This object doesn't support nations_wins_ranking method, " +
                "create one with nations wins ranking URL to call this" +
                "method.")
        return self._parse_regular_ranking_table(args, available_fields)

    def distance_ranking(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses ranking with riders ridden distances from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name:
            - rider_url:
            - team_name:
            - team_url:
            - rank: Rider's rank in the ranking.
            - nationality: Rider's nationality as 2 chars long country code.
            - distance: Rider's ridden distance in the season as KMs.

        :raises ExpectedParsingError: When the table from HTML is not a
            distance ranking table.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rider_name",
            "rider_url",
            "team_name",
            "team_url",
            "rank",
            "nationality",
            "distance",
        )
        if self._ranking_type() != "distance":
            raise ExpectedParsingError(
                "This object doesn't support distance_ranking method, " +
                "create one with distance ranking URL to call this" +
                "method.")
        fields = parse_table_fields_args(args, available_fields)
        casual_fields = [f for f in fields if f not in ("distance")]

        distance_ranking_table_html = self.html.css_first("span > table")
        table_parser = TableParser(distance_ranking_table_html)
        table_parser.parse(casual_fields)

        if "distance" in fields:
            distances = table_parser.parse_extra_column("KMs",
                lambda x: int(x) if x else 0)
            table_parser.extend_table("distance", distances)
        return table_parser.table

    def racedays_ranking(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses ranking with riders ridden racedays from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name:
            - rider_url:
            - team_name:
            - team_url:
            - rank: Rider's rank in the ranking.
            - nationality: Rider's nationality as 2 chars long country code.
            - racedays: Rider's ridden racedays in the season.

        :raises ExpectedParsingError: When the table from HTML is not a
            racedays ranking table.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rider_name",
            "rider_url",
            "team_name",
            "team_url",
            "rank",
            "nationality",
            "racedays",
        )
        if self._ranking_type() != "racedays":
            raise ExpectedParsingError(
                "This object doesn't support distance_ranking method, " +
                "create one with distance ranking URL to call this" +
                "method.")
        fields = parse_table_fields_args(args, available_fields)
        casual_fields = [f for f in fields if f not in ("racedays")]

        distance_ranking_table_html = self.html.css_first("span > table")
        table_parser = TableParser(distance_ranking_table_html)
        table_parser.parse(casual_fields)

        if "racedays" in fields:
            racedays = table_parser.parse_extra_column("Racedays",
                lambda x: int(x) if x else 0)
            table_parser.extend_table("racedays", racedays)
        return table_parser.table

    def dates_select(self) -> List[Dict[str, str]]:
        """
        Parses dates select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "date"))

    def nations_select(self) -> List[Dict[str, str]]:
        """
        Parses nations select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "nation"))

    def teams_select(self) -> List[Dict[str, str]]:
        """
        Parses teams select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "team"))

    def pages_select(self) -> List[Dict[str, str]]:
        """
        Parses pages select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "offset"))

    def teamlevels_select(self) -> List[Dict[str, str]]:
        """
        Parses team levels select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "teamlevel"))

    def _ranking_type(self) -> Literal["individual",
                                       "nations",
                                       "teams",
                                       "races",
                                       "distance",
                                       "racedays",
                                       "individual_wins",
                                       "team_wins",
                                       "nation_wins"]:
        """
        Finds out what is the ranking type of the object based on URL

        :return: Ranking type.
        """

        relative_url = self.relative_url()
        if len(relative_url.split("/")) < 3 and "?" not in relative_url:
            return "individual"
        if "races" in relative_url:
            l = [m.start() for m in re.finditer("races", relative_url)]
            if relative_url[l[0] - 1] != "-":
                return "races"
        if "distance" in relative_url:
            return "distance"
        if "racedays" in relative_url:
            return "racedays"
        if "wins-individual" in relative_url:
            return "individual_wins"
        if "wins-teams" in relative_url:
            return "team_wins"
        if "wins-nations" in relative_url:
            return "nation_wins"
        if "nations" in relative_url:
            return "nations"
        if "teams" in relative_url:
            return "teams"
        return "individual"

    def _parse_regular_ranking_table(self,
            args: Tuple[str, ...],
            available_fields: Tuple[str, ...]) -> List[Dict[str, Any]]:
        """
        Does general ranking parsing procedure using TableParser.

        :param args: Parsing method args (only the ones that
            `TableParser.parse` method is able to parse).
        :param available_fields: Available table fields for parsing method
        :return: Table with wanted fields.
        """
        fields = parse_table_fields_args(args, available_fields)
        html_table = self.html.css_first("table")
        table_parser = TableParser(html_table)
        table_parser.parse(fields)
        return table_parser.table



import calendar
from typing import Any, Dict, List

from .scraper import Scraper
from .table_parser import TableParser
from .utils import get_day_month, parse_table_fields_args


class Rider(Scraper):
    """
    Scraper for rider HTML page.

    Usage:

    >>> from procyclingstats import Rider
    >>> rider = Rider("rider/tadej-pogacar")
    >>> rider.birthdate()
    '1998-9-21'
    >>> rider.parse()
    {
        'birthdate': '1998-9-21',
        'height': 1.76,
        'name': 'Tadej  Pogačar',
        'nationality': 'SI',
        ...
    }
    """
    def birthdate(self) -> str:
        """
        Parses rider's birthdate from HTML.

        :return: birthday of the rider in ``YYYY-MM-DD`` format.
        """
        general_info_html = self.html.css_first(".rdr-info-cont")
        bd_string = general_info_html.text(separator=" ", deep=False)
        bd_list = [item for item in bd_string.split(" ") if item][:3]
        [day, str_month, year] = bd_list
        month = list(calendar.month_name).index(str_month)
        return f"{year}-{month}-{day}"

    def place_of_birth(self) -> str:
        """
        Parses rider's place of birth from HTML

        :return: rider's place of birth (town only).
        """
        # normal layout
        try:
            place_of_birth_html = self.html.css_first(
                ".rdr-info-cont > span > span > a")
            return place_of_birth_html.text()
        # special layout
        except AttributeError:
            place_of_birth_html = self.html.css_first(
                ".rdr-info-cont > span > span > span > a")
            return place_of_birth_html.text()

    def name(self) -> str:
        """
        Parses rider's name from HTML.

        :return: Rider's name.
        """
        return self.html.css_first(".page-title > .main > h1").text()

    def weight(self) -> float:
        """
        Parses rider's current weight from HTML.

        :return: Rider's weigth in kilograms.
        """
        # normal layout
        try:
            weight_html = self.html.css(".rdr-info-cont > span")[1]
            return float(weight_html.text().split(" ")[1])
        # special layout
        except (AttributeError, IndexError):
            weight_html = self.html.css(".rdr-info-cont > span > span")[1]
            return float(weight_html.text().split(" ")[1])

    def height(self) -> float:
        """
        Parses rider's height from HTML.

        :return: Rider's height in meters.
        """
        # normal layout
        try:
            height_html = self.html.css_first(".rdr-info-cont > span > span")
            return float(height_html.text().split(" ")[1])
        # special layout
        except (AttributeError, IndexError):
            height_html = self.html.css_first(
                ".rdr-info-cont > span > span > span")
            return float(height_html.text().split(" ")[1])

    def nationality(self) -> str:
        """
        Parses rider's nationality from HTML.

        :return: Rider's current nationality as 2 chars long country code in
            uppercase.
        """
        # normal layout
        nationality_html = self.html.css_first(".rdr-info-cont > .flag")
        if nationality_html is None:
        # special layout
            nationality_html = self.html.css_first(
                ".rdr-info-cont > span > span")
        flag_class = nationality_html.attributes['class']
        return flag_class.split(" ")[-1].upper() # type:ignore


    def teams_history(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses rider's team history throughout career.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - team_name:
            - team_url:
            - season:
            - class: Team's class, e.g. ``WT``.
            - since: First day for rider in current season in the team in
              ``MM-DD`` format, most of the time ``01-01``.
            - until: Last day for rider in current season in the team in
              ``MM-DD`` format, most of the time ``12-31``.

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "season",
            "since",
            "until",
            "team_name",
            "team_url",
            "class"
        )
        fields = parse_table_fields_args(args, available_fields)
        seasons_html_table = self.html.css_first("ul.list.rdr-teams")
        table_parser = TableParser(seasons_html_table)
        casual_fields = [f for f in fields
                         if f in ("season", "team_name", "team_url")]
        if casual_fields:
            table_parser.parse(casual_fields)
        # add classes for row validity checking
        classes = table_parser.parse_extra_column(2,
            lambda x: x.replace("(", "").replace(")", "").replace(" ", "")
            if x and "retired" not in x.lower() else None)
        table_parser.extend_table("class", classes)
        if "since" in fields:
            until_dates = table_parser.parse_extra_column(-2,
                lambda x: get_day_month(x) if "as from" in x else "01-01")
            table_parser.extend_table("since", until_dates)
        if "until" in fields:
            until_dates = table_parser.parse_extra_column(-2,
                lambda x: get_day_month(x) if "until" in x else "12-31")
            table_parser.extend_table("until", until_dates)

        table = [row for row in table_parser.table if row['class']]
        # remove class field if isn't needed
        if "class" not in fields:
            for row in table:
                row.pop("class")
        return table

    def points_per_season_history(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses rider's points per season history.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - season:
            - points: PCS points gained throughout the season.
            - rank: PCS ranking position after the season.

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "season",
            "points",
            "rank"
        )
        fields = parse_table_fields_args(args, available_fields)
        points_table_html = self.html.css_first("table.rdr-season-stats")
        table_parser = TableParser(points_table_html)
        table_parser.parse(fields)
        return table_parser.table

    def points_per_speciality(self) -> Dict[str, int]:
        """
        Parses rider's points per specialty from HTML.

        :return: Dict mapping rider's specialties and points gained.
            Dict keys: one_day_races, gc, time_trial, sprint, climber
        """
        specialty_html = self.html.css(".pps > ul > li > .pnt")
        pnts = [int(e.text()) for e in specialty_html]
        keys = ["one_day_races", "gc", "time_trial", "sprint", "climber"]
        return dict(zip(keys, pnts))



import inspect
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

import requests
from selectolax.parser import HTMLParser

from .errors import ExpectedParsingError


class Scraper:
    """Base class for all scraping classes."""
    BASE_URL: str = "https://www.procyclingstats.com/"

    _public_nonparsing_methods = (
        "update_html",
        "parse",
        "relative_url"
    )
    """Public methods that aren't called by `parse` method."""

    def __init__(self, url: str, html: Optional[str] = None,
                 update_html: bool = True) -> None:
        """
        Creates scraper object that is by default ready for HTML parsing. Call
        parsing methods to parse data from HTML.

        :param url: URL of procyclingstats page to parse. Either absolute or
            relative.
        :param html: HTML to be parsed from, defaults to None. When passing the
            parameter, set `update_html` to False to prevent overriding or
            making useless request.
        :param update_html: Whether to make request to given URL and update
            `self.html`. When False `self.update_html` method has to be called
            manually to make object ready for parsing. Defaults to True.

        :raises ValueError: When given HTML or HTML from given URL is invalid,
            e.g. 'Page not found' is contained in the HTML.
        """
        # validate given URL
        self._url = self._make_url_absolute(url)
        self._html = None
        if html:
            self._html = HTMLParser(html)
            if not self._html_valid():
                raise ValueError("Given HTML is invalid.")
            self._set_up_html()
        if update_html:
            self.update_html()
            if not self._html_valid():
                raise ValueError(
                    f"HTML from given URL is invalid: '{self.url}'")
            self._set_up_html()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(url='{self.url}')"

    @property
    def url(self) -> str:
        """Absolute URL from URL that was passed when constructing."""
        return self._url

    @property
    def html(self) -> HTMLParser:
        """
        HTML that is used for parsing.

        :raises AttributeError: when HTML is None
        """
        if self._html is None:
            raise AttributeError(
                "In order to access HTML, update it using " +
                "`self.update_html` method.")
        return self._html

    def relative_url(self) -> str:
        """
        Makes relative URL from absolute url (cuts `self.BASE_URL` from URL).

        :return: Relative URL.
        """
        return "/".join(self._url.split("/")[3:])

    def update_html(self) -> None:
        """
        Calls request to `self.url` and updates `self.html` to HTMLParser
        object created from returned HTML.
        """
        html_str = requests.get(self._url).text \
            # pylint: disable=missing-timeout
        self._html = HTMLParser(html_str)

    def parse(self,
            exceptions_to_ignore: Tuple[
            Type[Exception], ...] = (ExpectedParsingError,),
            none_when_unavailable: bool = True) -> Dict[str, Any]:
        """
        Creates JSON like dict with parsed data by calling all parsing methods.
        Keys in dict are methods names and values parsed data

        :param exceptions_to_ignore: Tuple of exceptions that should be
            ignored when raised by parsing methods. Defaults to
            ``(ExpectedParsingError,)``
        :param none_when_unavailable: Whether to set dict value to None when
            method raises ignored exception. When False the key value pair is
            skipped. Defaults to True.
        :return: Dict with parsing methods mapping to parsed data.
        """
        parsing_methods = self._parsing_methods()
        parsed_data = {}
        for method_name, method in parsing_methods:
            try:
                parsed_data[method_name] = method()
            except exceptions_to_ignore:
                if none_when_unavailable:
                    parsed_data[method_name] = None
        return parsed_data

    def _decompose_url(self) -> List[str]:
        """
        Splits relative URL to list of strings.

        :return: Splitted relative URL without empty strings.
        """
        splitted_url = self.relative_url().split("/")
        return [part for part in splitted_url if part]

    def _parsing_methods(self) -> List[Tuple[str, Callable]]:
        """
        Gets all parsing methods from a class. That are all public methods
        except of methods listed in `_public_nonparsing_methods`.

        :return: List of tuples parsing methods names and parsing methods.
        """
        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        parsing_methods = []
        for method_name, method in methods:
            if (method_name[0] != "_"
                and method_name not in self._public_nonparsing_methods):
                parsing_methods.append((method_name, method))
        return parsing_methods

    def _make_url_absolute(self, url: str) -> str:
        """
        Makes absolute URL from given url (adds `self.base_url` to URL if
        needed).

        :param url: URL to format.
        :return: Absolute URL.
        """
        if "https" not in url:
            if url[0] == "/":
                url = self.BASE_URL + url[1:]
            else:
                url = self.BASE_URL + url
        return url

    def _set_up_html(self):
        """
        Empty method that should be overridden by subclasses if it's needed to
        modify HTML before parsing.
        """

    def _html_valid(self) -> bool:
        """
        Checks whether given HTML is valid based on some known invalid formats
        of invalid HTMLs.

        :return: True if given HTML is valid, otherwise False.
        """
        try:
            page_title = self.html.css_first(".page-title > .main > h1").text()
            assert page_title != "Page not found"

            page_title2 = self.html.css_first("div.page-content > div").text()
            assert page_title2 != ("Due to technical difficulties this page " +
            "is temporarily unavailable.")

            page_title3 = self.html.css_first(
                ".page-title > .main > h1").text()
            assert page_title3 != "Start"
            return True
        except AssertionError:
            return False



from typing import Any, Optional


class ExpectedParsingError(Exception):
    """
    Exception to raise when data can't be parsed due to some known factor

    :param message: Exception message to print when raised.
    """

    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self) -> str:
        return self.message


class UnexpectedParsingError(Exception):
    """
    Exception to raise when data can't be parsed due to some unknown factor

    :param message: Exception message to print when raised.
    """

    def __init__(self,
                 message: str = "Unexpected parsing error occured") -> None:
        self.message = message

    def __str__(self) -> str:
        return self.message

class ParsedValueInvalidError(Exception):
    """
    Exception to raise when parsed value didn't pass some kind of validity
    check.

    :param message: Exception message to print when raised.
    """

    def __init__(self, value: Any,
                 custom_message: Optional[str] = None) -> None:
        self.value = value
        self.custom_message = custom_message

    def __str__(self) -> str:
        if self.custom_message:
            return self.custom_message
        return f"Parsed value '{self.value}' didn't pass validity check. "



import datetime
import math
import re
from typing import Any, Dict, List, Tuple, Union

from selectolax.parser import HTMLParser, Node

from .errors import ExpectedParsingError


# date and time manipulation functions
def get_day_month(str_with_date: str) -> str:
    """
    Gets day and month from string containing day/month or day-month.

    :param str_with_date: String with day and month separated by - or /.
    :raises ValueError: When string doesn't contain day and month in wanted
    format.
    :return: String in `MM-DD` format.
    """
    day, month = "", ""
    # loop through string and check whether next 5 characters are in wanted
    # date format `day/month` or `day-month`
    for i, _ in enumerate(str_with_date[:-4]):
        if str_with_date[i:i + 2].isnumeric() and \
                str_with_date[i + 3:i + 5].isnumeric():
            if str_with_date[i + 2] == "/":
                [day, month] = str_with_date[i:i + 5].split("/")
            elif str_with_date[i + 2] == "-":
                [day, month] = str_with_date[i:i + 5].split("-")
    if day.isnumeric() and month.isnumeric():
        return f"{month}-{day}"
    # day or month weren't numeric so given string doesn't contain date in
    # wanted format
    raise ValueError(
        "Given string doesn't contain day and month in wanted format")

def convert_date(date: str) -> str:
    """
    Converts given date to `YYYY-MM-DD` format.

    :param date: Date to convert, day, month and year have to be separated by
    spaces and month has to be in word form e.g. `30 July 2022`.
    :return: Date in `YYYY-MM-DD` format.
    """
    [day, month, year] = date.split(" ")
    month = datetime.datetime.strptime(month, "%B").month
    month = f"0{month}" if month < 10 else str(month)
    return "-".join([year, month, day])

def timedelta_to_time(tdelta: datetime.timedelta) -> str:
    """
    Converts timedelta object to time in `H:MM:SS` format.

    :param tdelta: Timedelta to convert.
    :return: Formatted time.
    """
    time = str(tdelta).split(" ")
    if len(time) > 1:
        days = time[0]
        time = time[2]
        hours = int(time.split(":")[0]) + (24 * int(days))
        minutes_seconds = ":".join(time.split(":")[1:])
    else:
        hours = time[0].split(":")[0]
        minutes_seconds = ":".join(time[0].split(":")[1:])
    return f"{hours}:{minutes_seconds}"

def time_to_timedelta(time: str) -> datetime.timedelta:
    """
    Converts time in `H:MM:SS` format to timedelta object.

    :param time: Time to convert.
    :return: Timedelta object.
    """
    [hours, minutes, seconds] = [int(value) for value in time.split(":")]
    return datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds)

def format_time(time: str) -> str:
    """
    Convert time from `M:SS` or `MM:SS` format to `H:MM:SS` format.

    :param time: Time to convert.
    :return: Formatted time e.g. `31:03:11`.
    """
    splitted_time = time.split(":")
    # make minutes and seconds two digits long
    for i, time_val in enumerate(splitted_time [-2:]):
        if len(time_val) == 1:
            splitted_time[i] = "0" + time_val
    time_str = ":".join(splitted_time)
    # add hours if needed
    if len(splitted_time) == 2:
        time_str = "0:" + time_str
    return time_str

def add_times(time1: str, time2: str) -> str:
    """
    Adds two given times with minutes and seconds or with hours optionally
    together.

    :param time1: Time separated with colons.
    :param time2: Time separated with colons.
    :return: Time in `H:MM:SS` format.
    """
    tdelta1 = time_to_timedelta(format_time(time1))
    tdelta2 = time_to_timedelta(format_time(time2))
    tdelta = tdelta1 + tdelta2
    return timedelta_to_time(tdelta)

# HTML parsing functions
def parse_select(select_menu: Node) -> List[Dict[str, str]]:
    """
    Parses select menu.

    :param select_menu: Select menu HTML.
    :return: Parsed select menu represented as list of dicts with keys `text`
    and `value`.
    """
    table = []
    for option in select_menu.css("option"):
        table.append({
            "text": option.text(),
            "value": option.attributes['value']
        })
    return table

def select_menu_by_name(html: Union[Node, HTMLParser], name_attr: str) -> Node:
    """
    Finds select menu my it's name attribute.

    :param html: HTML to find select menu in.
    :param name_attr: Name attribute of wanted select menu.
    :raises ExpectedParsingError: When select menu with given name attribute
    isn't contained in given HTML.
    :return: Wanted select menu HTML.
    """
    select_html = html.css_first(f"select[name={name_attr}]")
    if not select_html:
        raise ExpectedParsingError(f"'{name_attr}' select not in page HTML.")
    return select_html


# other functions
def join_tables(table1: List[Dict[str, Any]],
               table2: List[Dict[str, Any]],
               join_key: str,
               skip_missing: bool = False) -> List[Dict[str, Any]]:
    """
    Join given tables to one by joining rows which `join_key` values are
    matching.

    :param table1: Table represented as list of dicts where every row has
    `join_key`.
    :param table2: Table represented as list of dicts where every row has
    `join_key`.
    :param join_key: Field used for finding matching rows, e.g. `rider_url`.
    :param skip_missing: If set to False, error is raised when table1 and
        table2 don't have same join_keys. Otherwise only rows with join_keys
        present in both tables are added.
    :return: Tables joined together into one table.
    """
    table2_dict = {row[join_key]: row for row in table2}
    table = []
    for row in table1:
        if not skip_missing or table2_dict.get(row[join_key]):
            table.append({**table2_dict[row[join_key]], **row})
    return table

def parse_table_fields_args(args: Tuple[str],
                            available_fields: Tuple[str, ...]) -> List[str]:
    """
    Check whether given args are valid and get table fields.

    :param args: Args to be validated.
    :param available_fields: Args that would be valid.
    :raises ValueError: When one of args is not valid.
    :return: Table fields, args if any were given, otherwise all available
    fields.
    """
    for arg in args:
        if arg not in available_fields:
            raise ValueError("Invalid field argument")
    if args:
        return list(args)
    return list(available_fields)



from typing import Any, Dict, List, Literal, Optional

from selectolax.parser import HTMLParser, Node

from .errors import ExpectedParsingError
from .scraper import Scraper
from .table_parser import TableParser
from .utils import (add_times, convert_date, format_time, join_tables,
                    parse_table_fields_args)


class Stage(Scraper):
    """
    Scraper for stage results HTML page.

    Usage:

    >>> from procyclingstats import Stage
    >>> stage = Stage("race/tour-de-france/2022/stage-18")
    >>> stage.date()
    '2022-07-21'
    >>> stage.parse()
    {
        'arrival': Hautacam
        'date': '2022-07-21'
        'departure': 'Lourdes'
        'distance': 143.2
        'gc': [
            {
                'age': 25,
                'bonus': 32,
                'nationality': 'DK',
                'pcs_points': 0,
                'prev_rank': 1,
                'rank': 1,
                'rider_name': 'VINGEGAARD Jonas',
                'rider_url': 'rider/jonas-vingegaard-rasmussen',
                'team_name': 'Jumbo-Visma',
                'team_url': 'team/team-jumbo-visma-2022',
                'time': '71:53:34',
                'uci_points': 25.0
            },
            ...
        ],
        ...
    }
    """
    _tables_path = ".result-cont table"

    def _set_up_html(self) -> None:
        """
        Overrides Scraper method. Modifies HTML if stage is TTT by adding team
        ranks to riders.
        """
        # add team ranks to every rider's first td element, so it's possible
        # to map teams to riders based on their rank
        categories = self.html.css(self._tables_path)
        if not categories:
            return
        results_table_html = categories[0]
        if self.stage_type() != "TTT":
            return
        current_rank_node = None
        for column in results_table_html.css("tr > td:first-child"):
            rank = column.text()
            if rank:
                current_rank_node = column
            elif current_rank_node:
                column.replace_with(current_rank_node) # type: ignore

    def is_one_day_race(self) -> bool:
        """
        Parses whether race is an one day race from HTML.

        :return: Whether the race is an one day race.
        """
        # If there are elements with .restabs class (Stage/GC... menu), the race
        # is a stage race
        return len(self.html.css(".restabs")) == 0

    def distance(self) -> float:
        """
        Parses stage distance from HTML.

        :return: Stage distance in kms.
        """
        distance = self._stage_info_by_label("Distance")
        return float(distance.split(" km")[0])

    def profile_icon(self) -> Literal["p0", "p1", "p2", "p3", "p4", "p5"]:
        """
        Parses profile icon from HTML.

        :return: Profile icon e.g. ``p4``, the higher the number is the more
            difficult the profile is.
        """
        profile_html = self.html.css_first("span.icon")
        return profile_html.attributes['class'].split(" ")[2] # type: ignore

    def stage_type(self) -> Literal["ITT", "TTT", "RR"]:
        """
        Parses stage type from HTML.

        :return: Stage type, e.g. ``ITT``.
        """
        stage_name_html = self.html.css_first(".sub > .blue")
        stage_name2_html = self.html.css_first("div.main > h1")
        stage_name = stage_name_html.text()
        stage_name2 = stage_name2_html.text()
        if "ITT" in stage_name or "ITT" in stage_name2:
            return "ITT"
        if "TTT" in stage_name or "TTT" in stage_name2:
            return "TTT"
        return "RR"

    def winning_attack_length(self):
        return None

    def vertical_meters(self) -> Optional[int]:
        """
        Parses vertical meters gained throughout the stage from HTML.

        :return: Vertical meters.
        """
        vert_meters = self._stage_info_by_label("Vert. meters")
        if vert_meters:
            return int(vert_meters)
        return None

    def date(self) -> str:
        """
        Parses date when stage took place from HTML.

        :return: Date when stage took place in ``YYYY-MM-DD`` format.
        """
        date = self._stage_info_by_label("Date")
        return convert_date(date.split(", ")[0])

    def departure(self) -> str:
        """
        Parses departure of the stage from HTML.

        :return: Departure of the stage.
        """
        return self._stage_info_by_label("Departure")

    def arrival(self) -> str:
        """
        Parses arrival of the stage from HTML.

        :return: Arrival of the stage.
        """
        return self._stage_info_by_label("Arrival")

    def won_how(self) -> str:
        """
        Parses won how string from HTML.

        :return: Won how string e.g ``Sprint of small group``.
        """
        return self._stage_info_by_label("Won how")

    def race_startlist_quality_score(self) -> int:
        """
        Parses race startlist quality score from HTML.

        :return: Race startlist quality score.
        """
        return int(self._stage_info_by_label("Startlist quality score"))

    def profile_score(self) -> Optional[int]:
        """
        Parses profile score from HTML.

        :return: Profile score.
        """
        profile_score = self._stage_info_by_label("Profile")
        if profile_score:
            return int(profile_score)
        return None


    def pcs_points_scale(self) -> str:
        """
        Parses PCS points scale from HTML.

        :return: PCS points scale, e.g. ``GT.A.Stage``.
        """
        return self._stage_info_by_label("Points scale")

    def uci_points_scale(self) -> str:
        """
        Parses UCI points scale from HTML.

        :return: UCI points scale, e.g. ``UCI scale``. Empty string when not
            found.
        """
        scale_str = self._stage_info_by_label("UCI scale")
        if scale_str:
            return scale_str.split()[0]
        return scale_str

    def climbs(self, *args: str) -> List[Dict[str, str]]:
        """
        Parses listed climbs from the stage. When climbs aren't listed returns
        empty list.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - climb_name:
            - climb_url: URL of the location of the climb, NOT the climb itself

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "climb_name",
            "climb_url"
        )
        fields = parse_table_fields_args(args, available_fields)
        climbs_html = self.html.css_first("ul.list.circle")
        if climbs_html is None:
            return []

        table_parser = TableParser(climbs_html)
        table_parser.parse(fields)
        return table_parser.table

    def results(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses main results table from HTML. If results table is TTT one day
        race, fields `age` and `nationality` are set to None if are requested,
        because they aren't contained in the HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name:
            - rider_url:
            - rider_number:
            - team_name:
            - team_url:
            - rank: Rider's result in the stage.
            - status: ``DF``, ``DNF``, ``DNS``, ``OTL`` or ``DSQ``.
            - age: Rider's age.
            - nationality: Rider's nationality as 2 chars long country code.
            - time: Rider's time in the stage.
            - bonus: Bonus seconds that the rider gained in the stage.
            - pcs_points:
            - uci_points:

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rider_name",
            "rider_url",
            "rider_number",
            "team_name",
            "team_url",
            "rank",
            "status",
            "age",
            "nationality",
            "time",
            "bonus",
            "pcs_points",
            "uci_points"
        )
        fields = parse_table_fields_args(args, available_fields)
        # remove other result tables from html
        # because of one day races self._table_index isn't used here
        categories = self.html.css(self._tables_path)
        results_table_html = categories[0]
        # Results table is empty
        if (not results_table_html or
            not results_table_html.css_first("tbody > tr")):
            raise ExpectedParsingError("Results table not in page HTML")
        # parse TTT table
        if self.stage_type() == "TTT":
            table = self._ttt_results(results_table_html, fields)
            # set status of all riders to DF because status information isn't
            # contained in the HTML of TTT results
            if "status" in fields:
                for row in table:
                    row['status'] = "DF"
            # add extra elements from GC table if possible and needed
            gc_table_html = self._table_html("gc")
            if (not self.is_one_day_race() and gc_table_html and
                ("nationality" in fields or "age" in fields)):
                table_parser = TableParser(gc_table_html)
                extra_fields = [f for f in fields
                                if f in ("nationality", "age", "rider_url")]
                # add rider_url for table joining purposes
                extra_fields.append("rider_url")
                table_parser.parse(extra_fields)
                table = join_tables(table, table_parser.table, "rider_url",
                    True)
            elif "nationality" in fields or "age" in fields or \
                "rider_number" in fields:
                for row in table:
                    if "nationality" in fields:
                        row['nationality'] = None
                    if "age" in fields:
                        row['age'] = None
                    if "rider_number" in fields:
                        row['rider_number'] = None
            # remove rider_url from table if isn't needed
            if "rider_url" not in fields:
                for row in table:
                    row.pop("rider_url")
        else:
            table_parser = TableParser(results_table_html)
            table_parser.parse(fields)
            table = table_parser.table
        return table

    def gc(self, *args: str) -> List[Dict[str, Any]]: \
        # pylint: disable=invalid-name
        """
        Parses GC results table from HTML. When GC is unavailable, empty list
        is returned.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name:
            - rider_url:
            - rider_number:
            - team_name:
            - team_url:
            - rank: Rider's GC rank after the stage.
            - prev_rank: Rider's GC rank before the stage.
            - age: Rider's age.
            - nationality: Rider's nationality as 2 chars long country code.
            - time: Rider's GC time after the stage.
            - bonus: Bonus seconds that the rider gained throughout the race.
            - pcs_points:
            - uci_points:

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rider_name",
            "rider_url",
            "rider_number",
            "team_name",
            "team_url",
            "rank",
            "prev_rank",
            "age",
            "nationality",
            "time",
            "bonus",
            "pcs_points",
            "uci_points"
        )
        fields = parse_table_fields_args(args, available_fields)
        # remove other result tables from html
        gc_table_html = self._table_html("gc")
        if not gc_table_html:
            return []
        table_parser = TableParser(gc_table_html)
        table_parser.parse(fields)
        return table_parser.table

    def points(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses points classification results table from HTML. When points
        classif. is unavailable empty list is returned.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name:
            - rider_url:
            - rider_number:
            - team_name:
            - team_url:
            - rank: Rider's points classif. rank after the stage.
            - prev_rank: Rider's points classif. rank before the stage.
            - points: Rider's points classif. points after the stage.
            - age: Rider's age.
            - nationality: Rider's nationality as 2 chars long country code.
            - pcs_points:
            - uci_points:

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rider_name",
            "rider_url",
            "rider_number",
            "team_name",
            "team_url",
            "rank",
            "prev_rank",
            "points",
            "age",
            "nationality",
            "pcs_points",
            "uci_points"
        )
        fields = parse_table_fields_args(args, available_fields)
        # remove other result tables from html
        points_table_html = self._table_html("points")
        if not points_table_html:
            return []
        table_parser = TableParser(points_table_html)
        table_parser.parse(fields)
        return table_parser.table

    def kom(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses KOM classification results table from HTML. When KOM classif. is
        unavailable empty list is returned.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name:
            - rider_url:
            - rider_number:
            - team_name:
            - team_url:
            - rank: Rider's KOM classif. rank after the stage.
            - prev_rank: Rider's KOM classif. rank before the stage.
            - points: Rider's KOM points after the stage.
            - age: Rider's age.
            - nationality: Rider's nationality as 2 chars long country code.
            - pcs_points:
            - uci_points:

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rider_name",
            "rider_url",
            "rider_number",
            "team_name",
            "team_url",
            "rank",
            "prev_rank",
            "points",
            "age",
            "nationality",
            "pcs_points",
            "uci_points"
        )
        fields = parse_table_fields_args(args, available_fields)
        # remove other result tables from html
        kom_table_html = self._table_html("kom")
        if not kom_table_html:
            return []
        table_parser = TableParser(kom_table_html)
        table_parser.parse(fields)
        return table_parser.table

    def youth(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses youth classification results table from HTML. When youth classif
        is unavailable empty list is returned.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name:
            - rider_url:
            - rider_number:
            - team_name:
            - team_url:
            - rank: Rider's youth classif. rank after the stage.
            - prev_rank: Rider's youth classif. rank before the stage.
            - time: Rider's GC time after the stage.
            - age: Rider's age.
            - nationality: Rider's nationality as 2 chars long country code.
            - pcs_points:
            - uci_points:

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "rider_name",
            "rider_url",
            "rider_number",
            "team_name",
            "team_url",
            "rank",
            "prev_rank",
            "time",
            "age",
            "nationality",
            "pcs_points",
            "uci_points"
        )
        fields = parse_table_fields_args(args, available_fields)
        youth_table_html = self._table_html("youth")
        if not youth_table_html:
            return []
        table_parser = TableParser(youth_table_html)
        table_parser.parse(fields)
        return table_parser.table

    def teams(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses teams classification results table from HTML. When teams
        classif. is unavailable empty list is returned.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - team_name:
            - team_url:
            - rank: Teams's classif. rank after the stage.
            - prev_rank: Team's classif. rank before the stage.
            - time: Team's total GC time after the stage.
            - nationality: Team's nationality as 2 chars long country code.

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "team_name",
            "team_url",
            "rank",
            "prev_rank",
            "time",
            "nationality"
        )
        fields = parse_table_fields_args(args, available_fields)
        teams_table_html = self._table_html("teams")
        if not teams_table_html:
            return []
        table_parser = TableParser(teams_table_html)
        table_parser.parse(fields)
        return table_parser.table

    def _stage_info_by_label(self, label: str) -> str:
        """
        Finds infolist value for given label.

        :param label: Label to find value for.
        :return: Value of given label. Empty string when label is not in
            infolist.
        """
        for row in self.html.css("ul.infolist > li"):
            row_text = row.text(separator="\n").split("\n")
            row_text = [x for x in row_text if x != " "]
            if label in row_text[0]:
                if len(row_text) > 1:
                    return row_text[1]
                else:
                    return ""
        return ""

    def _table_html(self, table: Literal[
            "stage",
            "gc",
            "points",
            "kom",
            "youth",
            "teams"]) -> Optional[Node]:
        """
        Get HTML of a .result-cont table with results based on `table` param.

        :param table: Keyword of wanted table that occures in .restabs menu.
        :return: HTML of wanted HTML table, None when not found.
        """
        categories = self.html.css(".result-cont")
        for i, element in enumerate(self.html.css("ul.restabs > li > a")):
            if table in element.text().lower():
                return categories[i].css_first("table")
        return None

    @staticmethod
    def _ttt_results(results_table_html: Node,
                     fields: List[str]) -> List[Dict[str, Any]]:
        """
        Parses data from TTT results table.

        :param results_table_html: TTT results table HTML.
        :param fields: Fields that returned table should have. Available are
            all `results` table fields with the exception of age,
            nationality and rider_number.
        :return: Table with wanted fields.
        """
        team_fields = [
            "rank",
            "team_name",
            "team_url",
        ]
        rider_fields = [
            "rank",
            "rider_name",
            "rider_url",
            "pcs_points",
            "uci_points",
            "bonus"
        ]
        team_fields_to_parse = [f for f in team_fields if f in fields]
        rider_fields_to_parse = [f for f in rider_fields if f in fields]

        # add rank field to fields for joining tables purposes
        if "rank" not in fields:
            rider_fields_to_parse.append("rank")
            team_fields_to_parse.append("rank")
        # add rider_url for joining table with nationality or age from other
        # table, if isn't nedded is removed from table in self.results method
        if "rider_url" not in fields:
            rider_fields_to_parse.append("rider_url")

        # create two copies of HTML table (one for riders and one for teams),
        # so we won't modify self.html
        riders_elements = HTMLParser(results_table_html.html) # type: ignore
        riders_table = riders_elements.css_first("table")
        teams_elements = HTMLParser(results_table_html.html) # type: ignore
        teams_table = teams_elements.css_first("table")
        # remove unwanted rows from both tables
        riders_table.unwrap_tags(["tr.team"])
        teams_table.unwrap_tags(["tr:not(.team)"])
        teams_parser = TableParser(teams_table)
        teams_parser.parse(team_fields_to_parse)
        riders_parser = TableParser(riders_table)
        riders_parser.parse(rider_fields_to_parse)

        # add time of every rider to the table
        if "time" in fields:
            team_times = teams_parser.parse_extra_column("Time", format_time)
            # riders extra times from second HTML table column, if there is no
            # extra time, time is set to None
            riders_extra_times = riders_parser.parse_extra_column(1,
                lambda x: format_time(x.split("+")[1]) if
                len(x.split("+")) >= 2 else "0:00:00")

            riders_parser.extend_table("rider_time", riders_extra_times)
            teams_parser.extend_table("time", team_times)

            table = join_tables(riders_parser.table, teams_parser.table,
                "rank")
            # add team times and rider_extra times together and remove
            # rider_time field from table
            for row in table:
                rider_extra_time = row.pop('rider_time')
                row['time'] = add_times(row['time'], rider_extra_time)
        else:
            table = join_tables(riders_parser.table, teams_parser.table,
                "rank")
        # sort by name for consistent testing results (url is in fields by
        # default)
        table.sort(key = lambda x: x['rider_url'])
        # sort by rank to get default rank order
        table.sort(key = lambda x: x['rank'])
        if "rank" not in fields:
            for row in table:
                row.pop("rank")
        # for row in table:
        #     print(row['rider_url'])
        return table



from typing import Any, Dict, List

from .errors import ExpectedParsingError
from .scraper import Scraper
from .table_parser import TableParser
from .utils import parse_select, parse_table_fields_args, select_menu_by_name


class RiderResults(Scraper):
    """
    Scraper for rider results HTML page.

    Supported is besides of default results table also final 5k results table
    which can be parsed using ``final_n_km_results`` method.

    Usage:

    >>> from procyclingstats import RiderResults
    >>> rider_results = RiderResults("rider/alberto-contador/results/final-5k-analysis")
    >>> # for normal results table use "rider/alberto-contador/results" URL
    >>> rider_results.final_n_km_results()
    [
        {
            'average_percentage': 11.8,
            'class': '2.UWT',
            'date': '2017-09-09',
            'nationality': 'ES',
            'rank': 1,
            'stage_name': 'Vuelta a España | Stage 20',
            'stage_url': 'race/vuelta-a-espana/2017/stage-20',
            'vertical_meters': 590},
        },
        ...
    ]
    >>> rider_results.parse()
    {
        'categories_select': None,
        'final_n_km_results': [
            {
                'average_percentage': 11.8,
                'class': '2.UWT',
                'date': '2017-09-09',
                'nationality': 'ES',
                'rank': 1,
                'stage_name': 'Vuelta a España | Stage 20',
                'stage_url': 'race/vuelta-a-espana/2017/stage-20',
                'vertical_meters': 590},
            },
            ...
        ],
        'nations_select': None,
        ...
    }
    """
    def _html_valid(self) -> bool:
        """
        Extends Scraper method for validating HTMLs.

        :return: True if given HTML is valid, otherwise False
        """
        try:
            assert super()._html_valid()
            page_title = self.html.css_first(".page-content > h2").text()
            assert page_title in ("All results",
                "Top results final 5k analysis")
            return True
        except AssertionError:
            return False

    def _set_up_html(self):
        """Overrides Scraper method. Removes last table row with sum stats."""
        results_table_html = self.html.css_first("table")
        if not results_table_html:
            return
        for row in results_table_html.css("tr"):
            if "class" in row.attributes and row.attributes['class'] == "sum":
                row.decompose()

    def results(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses general rider's results table from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - stage_url:
            - stage_name:
            - distance:
            - nationality: Nationality of the stage race.
            - date: Date when the stage occured in ``YYYY-MM-DD`` format.
            - rank: Rider's result in the stage.
            - class: Class of the stage's race, e.g. ``2.UWT``.
            - pcs_points:
            - uci_points:

        :raises ExpecterParsingError: When the table from HTML isn't a results
            table.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "date",
            "rank",
            "stage_url",
            "stage_name",
            "nationality",
            "class",
            "distance",
            "pcs_points",
            "uci_points"
        )
        if self.html.css_first(".page-content > h2").text() != "All results":
            error_msg = ("This object doesn't support 'results' method. " +
                "Create one from rider's default results table to call this " +
                "method")
            raise ExpectedParsingError(error_msg)

        fields = parse_table_fields_args(args, available_fields)
        results_table_html = self.html.css_first("table")
        table_parser = TableParser(results_table_html)
        table_parser.parse(fields)
        return table_parser.table

    def final_n_km_results(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses rider's final n KMs results table from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - stage_url:
            - stage_name:
            - nationality: Nationality of the stage race.
            - date: Date when the stage occured in ``YYYY-MM-DD`` format.
            - rank: Rider's result in the stage.
            - class: Class of the stage's race, e.g. ``2.UWT``.
            - vertical_meters: Vertical meters gained in final n KMs.
            - average_percentage: Average percentage of last n KMs.

        :raises ExpecterParsingError: When the table from HTML isn't a final n
            KMs results table.
        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "date",
            "rank",
            "stage_url",
            "stage_name",
            "nationality",
            "class",
            "vertical_meters",
            "average_percentage"
        )
        if (self.html.css_first(".page-content > h2").text() !=
                "Top results final 5k analysis"):
            error_msg = ("This object doesn't support 'final_n_km_results'" +
            "method. Create one from rider's final n km results table to " +
            " call this  method")
            raise ExpectedParsingError(error_msg)

        fields = parse_table_fields_args(args, available_fields)
        casual_fields = [f for f in fields
            if f not in ("vertical_meters", "average_percentage")]

        results_table_html = self.html.css_first("div:nth-child(4) table")
        table_parser = TableParser(results_table_html)
        table_parser.parse(casual_fields)
        # add vertical meters column if needed
        if "vertical_meters" in fields:
            vms = table_parser.parse_extra_column("Vertical meters", int)
            table_parser.extend_table("vertical_meters", vms)
        # add average percentages column if needed
        if "average_percentage" in fields:
            percentages = table_parser.parse_extra_column("Avg. %", float)
            table_parser.extend_table("average_percentage", percentages)
        return table_parser.table

    def seasons_select(self) -> List[Dict[str, str]]:
        """
        Parses seasons select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "xseason"))

    def races_select(self) -> List[Dict[str, str]]:
        """
        Parses race select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "race"))

    def pages_select(self) -> List[Dict[str, str]]:
        """
        Parses race select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "offset"))

    def stage_types_select(self) -> List[Dict[str, str]]:
        """
        Parses stage types select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "type"))

    def nations_select(self) -> List[Dict[str, str]]:
        """
        Parses nations select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "znation"))

    def categories_select(self) -> List[Dict[str, str]]:
        """
        Parses categories select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        return parse_select(select_menu_by_name(self.html, "category"))



import os
import sys

from .race_climbs_scraper import RaceClimbs
from .race_scraper import Race
from .race_startlist_scraper import RaceStartlist
from .ranking_scraper import Ranking
from .rider_results_scraper import RiderResults
from .rider_scraper import Rider
from .scraper import Scraper
from .stage_scraper import Stage
from .team_scraper import Team

__all__ = [
    "Scraper",
    "RaceClimbs",
    "Race",
    "RaceStartlist",
    "Ranking",
    "RiderResults",
    "Rider",
    "Stage",
    "Team"
]

sys.path.append(os.path.dirname(os.path.realpath(__file__)))



from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

from selectolax.parser import Node

from .errors import ExpectedParsingError, UnexpectedParsingError
from .utils import add_times, format_time


class TableParser:
    """
    Parser for HTML tables. Parsed content is stored in `self.table`, which is
    represented as list of dicts.

    :param html_table: HTML table to be parsed from.
    """

    table_row_dict: Dict[str, str] = {
        "tbody": "tr",
        "table": "tr",
        "ul": "li"
    }
    """Finds out what is the table row tag."""
    row_column_tag_dict: Dict[str, str] = {
        "tr": "td",
        "li": "div"
    }
    """Finds out what is the table row column tag."""

    def __init__(self, html_table: Node) -> None:
        self.table = []
        table_body = html_table.css_first("tbody")
        if table_body:
            self.html_table = table_body
            self.header = html_table.css_first("thead")
        else:
            self.html_table = html_table
            self.header = None

        self.table_row_tag = self.table_row_dict[self.html_table.tag]
        self.row_column_tag = self.row_column_tag_dict[self.table_row_tag]

        self.a_elements = self.html_table.css("a")
        self.table_length = len(self.html_table.css(self.table_row_tag))
        self.row_length = len(self.html_table.css(
            f"{self.table_row_tag}:first-child > {self.row_column_tag}"))

    def parse(self, fields: Union[List[str], Tuple[str, ...]]) -> None:
        """
        Parses HTML table to `self.table` (list of dicts) by calling given
        table parsing methods. Every parsed table row is dictionary with
        `fields` keys.

        :param fields: Table parsing methods of this class.
        :raises UnexpectedParsingError: When parsed field values aren't the
        same size as table length.

        :regular fields options:
            - rider_url
            - rider_name
            - team_url
            - team_name
            - stage_url
            - stage_name
            - nation_url
            - nation_name
            - age
            - nationality
            - time
            - bonus
            - profile_icon
            - season
            - rider_number

        :fields options for tables with a header:
            - rank
            - status
            - prev_rank
            - pcs_points
            - uci_points
            - points
            - class
            - first_places
            - second_places
            - third_places
            - distance
            - date
        """
        raw_table = []
        for _ in range(self.table_length):
            raw_table.append({})

        for field in fields:
            if field != "class":
                parsed_field_list = getattr(self, field)()
            # special case when field is called class
            else:
                parsed_field_list = getattr(self, "class_")()
            # field wasn't found in every table row, so isn't matching table
            # rows correctly
            if len(parsed_field_list) != self.table_length:
                message = f"Field '{field}' wasn't parsed correctly"
                raise UnexpectedParsingError(message)

            for row, parsed_value in zip(raw_table, parsed_field_list):
                row[field] = parsed_value

        # remove unwanted rows
        for row in raw_table:
            self.table.append(row)

        if "time" in fields and self.table:
            self._make_times_absolute()

    def extend_table(self, field_name: str, values: List[Any]):
        """
        Add given values to table.

        :param field_name: Name for column that's being added.
        :param values: Values which are being added.
        :raises ValueError: When values to add aren't the same length as table.
        """
        if len(values) != self.table_length:
            raise ValueError(
                "Given values has to be the same length as table rows count")
        if self.table:
            for row, value in zip(self.table, values):
                row[field_name] = value
        else:
            for value in values:
                self.table.append({field_name: value})

    def parse_extra_column(self, index_or_header_value: Union[int, str],
                     func: Callable = int,
                     separator: str = "") -> List[Any]:
        """
        Parses values from given column.

        :param index_or_header_value: Either index of column to parse (negative
        indexing works too) or column name from table header (table has to have
        a header in that case).
        :param func: Function to call on parsed text value, defaults to int.
        :param separator: Separator for text attributes given to `func`.
        :return: List with parsed values.
        """
        if isinstance(index_or_header_value, str):
            index = self._get_column_index_from_header(index_or_header_value)
        else:
            index = index_or_header_value
        if index < 0:
            index = self.row_length + index
        elements = self.html_table.css(
        f"{self.table_row_tag} > {self.row_column_tag}:nth-child({index+1})"
        )

        values = []
        for element in elements:
            values.append(func(element.text(separator=separator)))
        return values

    def rider_url(self) -> List[str]:
        return self._filter_a_elements("rider", True)

    def rider_name(self) -> List[str]:
        return self._filter_a_elements("rider", False)

    def team_url(self) -> List[str]:
        return self._filter_a_elements("team", True,
            lambda x: True if x.text() != "view" else False)

    def team_name(self) -> List[str]:
        return self._filter_a_elements("team", False,
            lambda x: True if x.text() != "view" else False)

    def stage_url(self) -> List[str]:
        return self._filter_a_elements("race", True)

    def stage_name(self) -> List[str]:
        return self._filter_a_elements("race", False)

    def nation_url(self) -> List[str]:
        nations_urls = self._filter_a_elements("nation", True)
        # return only urls to nation overview, not `pcs-season-wins`
        return [url for url in nations_urls if "pcs" not in url]

    def nation_name(self) -> List[str]:
        nations_texts = self._filter_a_elements("nation", False)
        # return text only when is not numeric, so doesn't represent number of
        # wins of the nation
        return [text for text in nations_texts
                if not text.isnumeric() and text != "-"]

    def climb_url(self) -> List[str]:
        """
        Parses all location elements hrefs from HTML. NOT only climbs, but for
        ease of use method is called climb name.

        :return: List of all climb URLs from table.
        """
        return self._filter_a_elements("location", True)

    def climb_name(self) -> List[str]:
        """
        Parses all location elements text values from HTML. NOT only climbs,
        but for ease of use method is called climb name.

        :return: List of all climb names from table.
        """
        return self._filter_a_elements("location", False)

    def age(self) -> List[int]:
        ages_elements = self.html_table.css(".age")
        return [int(age_e.text()) for age_e in ages_elements]

    def nationality(self) -> List[str]:
        flags_elements = self.html_table.css(".flag")
        flags = []
        for flag_e in flags_elements:
            if (flag_e.attributes['class'] and
                    " " in flag_e.attributes['class']):
                flags.append(flag_e.attributes['class'].split(" ")[1].upper())
        return flags

    def time(self) -> List[Optional[str]]:
        times_elements = self.html_table.css(".time")
        times = []
        for time_e in times_elements:
            time_e_text = time_e.text(separator="\n")
            rider_time = None
            for time_line in time_e_text.split("\n"):
                if ",," not in time_line and "″" not in time_line:
                    rider_time = time_line
                    break
            if rider_time == "-" or rider_time is None:
                rider_time = None
            else:
                rider_time = format_time(rider_time.replace(" ", ""))
            times.append(rider_time)
        return times

    def bonus(self) -> List[int]:
        """
        Parses all bonuses elements from the table. If there aren't any returns
        where every row has bonus 0.

        :return: List of bonuses.
        """
        bonuses_elements = self.html_table.css(".bonis")
        bonuses = []
        for bonus_e in bonuses_elements:
            bonus = bonus_e.text().replace("″", "")
            if not bonus:
                bonus = 0
            else:
                bonus = int(bonus)
            bonuses.append(bonus)
        if not bonuses:
            bonuses = [0 for _ in range(self.table_length)]
        return bonuses

    def profile_icon(self) -> List[Literal[
        "p0", "p1", "p2", "p3", "p4", "p5"
    ]]:
        icons_elements = self.html_table.css(".icon.profile")
        profiles = []
        for icon_e in icons_elements:
            classes = icon_e.attributes['class']
            if classes and len(classes.split(" ")) >= 3:
                profiles.append(classes.split(" ")[-1])
        return profiles

    def season(self) -> List[Optional[int]]:
        """
        Parses all season elements text values from table. If value is not
        numeric sesaon is set to None.

        :return: List of seasons.
        """
        seasons_elements = self.html_table.css(".season")
        seasons = []
        for season_e in seasons_elements:
            season_e_text = season_e.text()
            if season_e_text.isnumeric():
                seasons.append(int(season_e_text))
            else:
                seasons.append(None)
        return seasons
    
    def rider_number(self) -> List[Optional[int]]:
        bibs_elements = self.html_table.css(".bibs")
        return [int(bib_e.text()) if bib_e.text().isnumeric() else None \
            for bib_e in bibs_elements]

    def rank(self) -> List[Optional[int]]:
        possible_columns = ["Rnk", "pos", "Result", "#"]
        for column_name in possible_columns:
            try:
                return self.parse_extra_column(column_name,
                    lambda x: int(x) if x.isnumeric() else None)
            except ValueError:
                pass
        raise ValueError("Rank column wasn't found.")

    def status(self) -> List[Literal[
        "DF", "DNF", "DNS", "OTL", "DSQ"
    ]]:
        return self.parse_extra_column("Rnk",
            lambda x: "DF" if x.isnumeric() else x)

    def prev_rank(self) -> List[Optional[int]]:
        try:
            return self.parse_extra_column("Prev",
                lambda x: int(x) if x else None)
        except ValueError:
            return [None for _ in range(self.table_length)]

    def uci_points(self) -> List[Optional[float]]:
        try:
            return self.parse_extra_column("UCI",
                lambda x: float(x) if x and x.replace('.', '', 1).isdigit() else 0)
        except ValueError:
            return [0 for _ in range(self.table_length)]

    def pcs_points(self) -> List[Optional[int]]:
        try:
            return self.parse_extra_column("Pnt", lambda x: int(x) if x else 0)
        except ValueError:
            try:
                return self.parse_extra_column("PCS points",
                    lambda x: int(x) if x  and x.isdigit() else 0)
            except ValueError:
                return [0 for _ in range(self.table_length)]

    def points(self) -> List[int]:
        return self.parse_extra_column("Points", lambda x:
            float(x) if x else 0)

    def class_(self) -> List[str]:
        """
        Parses classes from table with a header. Method is called class_ so
        it won't be interchanged with class keyword. In parsed table underscore
        is removed.

        :return: List of classes.
        """
        return self.parse_extra_column("Class", str)

    def first_places(self) -> List[Optional[int]]:
        return self.parse_extra_column("Wins", lambda x:
            int(x) if x.isnumeric() else 0)

    def second_places(self) -> List[Optional[int]]:
        return self.parse_extra_column("2nd", lambda x:
            int(x) if x.isnumeric() else 0)

    def third_places(self) -> List[Optional[int]]:
        return self.parse_extra_column("3rd", lambda x:
            int(x) if x.isnumeric() else 0)

    def distance(self) -> List[float]:
        return self.parse_extra_column("KMs", lambda x:
            float(x) if x else None)

    def date(self) -> List[str]:
        return self.parse_extra_column("Date", str)

    def rename_field(self, field_name: str, new_field_name: str) -> None:
        """
        Renames field from table.

        :param field_name: Original field name.
        :param new_field_name: New name of original field.
        """
        for row in self.table:
            value = row.pop(field_name)
            row[new_field_name] = value

    def _get_column_index_from_header(self, column_name: str) -> int:
        if self.header is None:
            raise ExpectedParsingError(
                f"Can not parse '{column_name}' column without table header")
        for i, column_name_e in enumerate(self.header.css("th")):
            if column_name.lower() in column_name_e.text().lower():
                return i
        raise ValueError(
            f"'{column_name}' column isn't in table header")

    def _make_times_absolute(self, time_field: str = "time") -> None:
        """
        Sums all times from table with first time from table. Table has to have
        at least 2 rows.

        :param time_field: Field which represents wanted time, defaults to
        `time`.
        """
        first_time = self.table[0][time_field]
        for i, row in enumerate(self.table[1:]):
            if row[time_field]:
                try:
                    row[time_field] = add_times(first_time, row['time'])
                # if time is in invalid format
                except Exception:
                    if i == 0:
                        row[time_field] = "0:00:00"
                    # set the same time as previous rider
                    else:
                        row[time_field] = self.table[1:][i - 1]['time']


    def _filter_a_elements(self, keyword: str, get_href: bool,
            validator: Callable = lambda x: True) -> List[str]:
        """
        Filters from all a elements these which has at the beggining of their
        href given keyword and gets their href or text.

        :param keyword: Keyword that element's href should have.
        :param get_href: Whether to return the href of a element, when False
        text is returned.
        :param validator: Function to call on every a element. When returns
        True element is added to result list, otherwise not.
        :return: List of all a elements texts or hrefs with given keyword.
        """
        filtered_values = []
        for a_element in self.a_elements:
            href = a_element.attributes['href']
            if href and href.split("/")[0] == keyword and validator(a_element):
                if get_href:
                    filtered_values.append(href)
                else:
                    filtered_values.append(a_element.text())
        return filtered_values



from setuptools import setup

with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="procyclingstats",
    version="0.1.8",
    license="MIT",
    description="A Python API wrapper for procyclingstats.com",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Martin Madzin",
    author_email="madzin.m@gmail.com",
    packages=["procyclingstats"],
    url="https://github.com/themm1/procyclingstats",
    keywords="cycling cycling-stats procyclingstats scraper html-parsing" +
        " sports-analytics",
    install_requires=[
        "requests",
        "selectolax"
    ],
)



from typing import Any, Dict, List, Optional

from .scraper import Scraper
from .table_parser import TableParser
from .utils import (get_day_month, join_tables, parse_select,
                    parse_table_fields_args)


class Team(Scraper):
    """
    Scraper for team HTML page.

    Usage:

    >>> from procyclingstats import Team
    >>> team = Team("team/bora-hansgrohe-2022")
    >>> team.abbreviation()
    'BOH'
    >>> team.parse()
    {
        'abbreviation': 'BOH',
        'bike': 'Specialized',
        'history_select': [
            {
                'text': '2027 | BORA - hansgrohe',
                'value': 'team/bora-hansgrohe-2027/overview/'
            },
            ...
        ],
        'name': 'BORA - hansgrohe',
        ...
    }
    """
    def name(self) -> str:
        """
        Parses team display name from HTML.

        :return: Display name, e.g. ``BORA - hansgrohe``.
        """
        display_name_html = self.html.css_first(".page-title > .main > h1")
        return display_name_html.text().split(" (")[0]

    def nationality(self) -> str:
        """
        Parses team's nationality from HTML.

        :return: Team's nationality as 2 chars long country code in uppercase.
        """
        nationality_html = self.html.css_first(
            ".page-title > .main > span.flag")
        flag_class = nationality_html.attributes['class']
        return flag_class.split(" ")[1].upper() # type: ignore

    def status(self) -> str:
        """
        Parses team status (class) from HTML.

        :return: Team status as 2 chars long code in uppercase, e.g. ``WT``.
        """
        team_status_html = self.html.css_first(
            "div > ul.infolist > li:nth-child(1) > div:nth-child(2)")
        return team_status_html.text()

    def abbreviation(self) -> str:
        """
        Parses team abbreviation from HTML.

        :return: Team abbreviation as 3 chars long code in uppercase, e.g.
            ``BOH``
        """
        abbreviation_html = self.html.css_first(
            "div > ul.infolist > li:nth-child(2) > div:nth-child(2)")
        return abbreviation_html.text()

    def bike(self) -> str:
        """
        Parses team's bike brand from HTML.

        :return: Bike brand e.g. ``Specialized``.
        """
        bike_html = self.html.css_first(
            "div > ul.infolist > li:nth-child(3) > div:nth-child(2)")
        return bike_html.text()

    def wins_count(self) -> int:
        """
        Parses count of wins in corresponding season from HTML.

        :return: Count of wins in corresponding season.
        """
        wins_count_html = self.html.css_first(".team-kpi > li.nr")
        return int(wins_count_html.text())
    
    def pcs_points(self) -> Optional[int]:
        """
        Parses team's PCS points from HTML.

        :return: PCS points gained throughout corresponding year.
        """
        team_ranking_html = self.html.css_first(
            ".team-kpi > li.nr:nth-child(4)")
        if team_ranking_html.text().isnumeric():
            return int(team_ranking_html.text())
        else:
            return None
   

    def pcs_ranking_position(self) -> Optional[int]:
        """
        Parses team's PCS ranking position from HTML.

        :return: PCS team ranking position in corresponding year.
        """
        team_ranking_html = self.html.css_first(
            ".team-kpi > li.nr:nth-child(6)")
        if team_ranking_html.text().isnumeric():
            return int(team_ranking_html.text())
        else:
            return None
        
    def uci_ranking_position(self) -> Optional[int]:
        """
        Parses team's UCI ranking position from HTML.

        :return: UCI team ranking position in corresponding year.
        """
        team_ranking_html = self.html.css_first(
            ".team-kpi > li.nr:nth-child(8)")
        if team_ranking_html.text().isnumeric():
            return int(team_ranking_html.text())
        else:
            return None
   

    def history_select(self) -> List[Dict[str, str]]:
        """
        Parses team seasons select menu from HTML.

        :return: Parsed select menu represented as list of dicts with keys
            ``text`` and ``value``.
        """
        team_seasons_select_html = self.html.css_first("form > select")
        return parse_select(team_seasons_select_html)

    def riders(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses team riders in curresponding season from HTML.

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - rider_name
            - rider_url
            - nationality: Rider's nationality as 2 chars long country code.
            - age: Rider's age.
            - since: First rider's day in the team in corresponding season in
              ``MM-DD`` format, most of the time ``01-01``.
            - until: Last rider's day in the team in corresponding season in
              ``MM-DD`` format, most of the time ``12-31``.
            - career_points: Current riders's career points.
            - ranking_points: Current rider's points in PCS ranking.
            - ranking_position: Current rider's position in PCS ranking.

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "nationality",
            "rider_name",
            "rider_url",
            "age",
            "since",
            "until",
            "career_points",
            "ranking_points",
            "ranking_position"
        )
        casual_fields = [
            "nationality",
            "rider_name",
            "rider_url"]
        mapping = {}
        for i, li in enumerate(self.html.css("ul.riderlistTabs > li")):
            mapping[li.text()] = i
        all_tables = self.html.css("div.ridersTab")

        fields = parse_table_fields_args(args, available_fields)
        career_points_table_html = all_tables[mapping["points"]]
        table_parser = TableParser(career_points_table_html)
        career_points_fields = [field for field in fields
                         if field in casual_fields]
        # add rider_url to the table for table joining purposes
        if "rider_url" not in career_points_fields:
            career_points_fields.append("rider_url")
        table_parser.parse(career_points_fields)
        if "career_points" in fields:
            career_points = table_parser.parse_extra_column(2,
                lambda x: int(x) if x.isnumeric() else 0)
            table_parser.extend_table("career_points", career_points)
        table = table_parser.table

        # add ages to the table if needed
        if "age" in fields:
            ages_table_html = all_tables[mapping["age"]]
            ages_tp = TableParser(ages_table_html)
            ages_tp.parse(["rider_url"])
            ages = ages_tp.parse_extra_column(2, lambda x: int(x[:2]))
            ages_tp.extend_table("age", ages)
            table = join_tables(table, ages_tp.table, "rider_url")

        # add ranking points and positions to the table if needed
        if "ranking_position" in fields or "ranking_points" in fields:
            ranking_table_html = all_tables[mapping["ranking"]]
            ranking_tp = TableParser(ranking_table_html)
            ranking_tp.parse(["rider_url"])
            if "ranking_points" in fields:
                points = ranking_tp.parse_extra_column(2,
                    lambda x: int(x.replace("(", "").replace(")", ""))
                    if x.replace("(", "").replace(")", "").isnumeric() else 0)
                ranking_tp.extend_table("ranking_points", points)
            if "ranking_position" in fields:
                positions = ranking_tp.parse_extra_column(3,
                    lambda x: int(x) if x.isnumeric() else None)
                ranking_tp.extend_table("ranking_position", positions)
            table = join_tables(table, ranking_tp.table, "rider_url")

        # add rider's since and until dates to the table if needed
        if "since" in fields or "until" in fields:
            since_until_html_table = all_tables[mapping["name"]]
            since_tp = TableParser(since_until_html_table)
            since_tp.parse(["rider_url"])
            if "since" in fields:
                since_dates = since_tp.parse_extra_column(2,
                    lambda x: get_day_month(x) if "as from" in x else "01-01")
                since_tp.extend_table("since", since_dates)
            if "until" in fields:
                until_dates = since_tp.parse_extra_column(2,
                    lambda x: get_day_month(x) if "until" in x else "12-31")
                since_tp.extend_table("until", until_dates)
            table = join_tables(table, since_tp.table, "rider_url")

        # remove rider_url field if wasn't requested and was used for joining
        # tables only
        if "rider_url" not in fields:
            for row in table:
                row.pop("rider_url")
        return table



import argparse
import sys
from pprint import pprint  # pylint: disable=unused-import
from typing import Any, Dict, List, Tuple, Type

from tabulate import tabulate

# imports all scraping classes that are listed in `scraper_classes` tuple and
# Scraper class
from .__init__ import (Race, RaceClimbs, RaceStartlist, Ranking, Rider,
                       RiderResults, Scraper, Stage, Team)

scraper_classes = (
    Race,
    RaceStartlist,
    Ranking,
    Rider,
    Stage,
    Team,
    RiderResults,
    RaceClimbs
)

def configure_parser():
    parser = argparse.ArgumentParser(
        prog="python -m procyclingstats",
        description=
        ("CLI to procyclingstats package. Scraper object to parse " +
        "given URL is evaluated automatically. When no scraper object is " +
        "able to parse given URL ValueError is raised. When ran in " +
        "interactive mode, nothing is printed and scraper object ready " +
        "for parsing is available as `obj`.")
    )
    parser.add_argument("url", metavar="url", type=str,
                        help="Absolute or relative URL of PCS page to parse.")
    parser.add_argument("--fulltable", action="store_const", const=True,
            default=False,
            help="Whether to print full or shortened tables in output.")
    return parser

def get_corresponding_scraping_class(relative_url: str) -> Any:
    """
    Returns scraping class for given URL (!!!does not work 100% of times!!!).

    :param relative_url: Relative URL of some PCS page.
    :return: Scraping class for the URL. None when not found.
    """
    splitted_url = relative_url.split("/")
    if (splitted_url[0] == "rider" and "results" in splitted_url) or \
        (relative_url[:9] == "rider.php"):  
        return RiderResults
    elif splitted_url[0] == "rider":
        return Rider
    elif len(splitted_url) >= 4 and splitted_url[0] == "race" and \
        ("stage" in splitted_url[3] or "gc" in splitted_url[3] or "prologue" \
        in splitted_url[3] or "result" in splitted_url):
        return Stage
    elif "rankings" in relative_url:
        return Ranking
    elif splitted_url[0] == "race" and "startlist" in splitted_url:
        return RaceStartlist
    elif "team" == splitted_url[0]:
        return Team
    elif splitted_url[0] == "race" and "climbs" in splitted_url:
        return RaceClimbs
    elif splitted_url[0] == "race":
        return Race
    return None

def run(args: argparse.Namespace) -> Scraper:
    """
    Runs CLI script with given arguments.

    :param args: Argparse arguments (currently 'url' and 'fulltable').
    :return: Scraper object created from given URL.
    """
    scraper_class = get_corresponding_scraping_class(args.url)
    scraper_obj = scraper_class(args.url)

    # object created, so return when running in interactive mode
    if sys.flags.interactive:
        print(f"Scraper object `{scraper_obj}` can now be accessed as `obj`.")
        return scraper_obj

    tables = {}
    # print basic one line data
    for key, value in scraper_obj.parse().items():
        if isinstance(value, list) and value:
            tables[key] = value
        else:
            print(key + ": " + str(value))

    # print tables
    for key, value in tables.items():
        print()
        print(key + ":")
        # tabulate full table
        if args.fulltable or len(value) <= 13:
            print(tabulate(value, headers="keys"))
        # tabulate shortened table (first and last 5 rows of table with dots
        # inbetween
        else:
            shortened_table = value[:5]
            for _ in range(3):
                shortened_table.append({})
                for key in shortened_table[0].keys():
                    shortened_table[-1][key] = "..."
            shortened_table.extend(value[-5:])
            print(tabulate(shortened_table, headers="keys"))
    return scraper_obj

def tab(table: List[Dict[str, Any]]) -> None:
    """
    Utility function for easier table tabulating when running in interactive
    mode.

    :param table: Table to tabulate.
    """
    print(tabulate(table, headers="keys"))

if __name__ == "__main__":
    arguments = configure_parser().parse_args()
    obj = run(arguments)



from typing import Any, Dict, List

from .errors import ExpectedParsingError
from .scraper import Scraper
from .table_parser import TableParser
from .utils import parse_table_fields_args


class RaceClimbs(Scraper):
    """
    Scraper for race climbs HTML page.

    Usage:

    >>> from procyclingstats import RaceClimbs
    >>> race_climbs = RaceClimbs("race/tour-de-france/2022/route/climbs")
    >>> race_climbs.climbs()
    [
        {
            'climb_name': "Côte d'Asnæs Indelukke",
            'climb_url': 'location/cote-d-asnaes-indelukke',
            'km_before_finnish': 140,
            'length': 1.1,
            'steepness': 5.3,
            'top': 63
        },
        ...
    ]
    >>> race_climbs.parse()
    {
        'climbs': [
            {
                'climb_name': "	Côte d'Asnæs Indelukke",
                'climb_url': 'location/cote-d-asnaes-indelukke',
                'km_before_finnish': 140,
                'length': 1.1,
                'steepness': 5.3,
                'top': 63
            },
            ...
        ]
    }

    """
    def _html_valid(self) -> bool:
        """
        Extends Scraper method for validating HTMLs.

        :return: True if given HTML is valid, otherwise False
        """
        return self.html.css_first("div.page-content > h2").text() == "Climbs"

    def climbs(self, *args: str) -> List[Dict[str, Any]]:
        """
        Parses race's climbs table from HTML. Note that not allways all info
        about the climbs is present (usually in older races).

        :param args: Fields that should be contained in returned table. When
            no args are passed, all fields are parsed.

            - climb_name:
            - climb_url: URL of the location of the climb, NOT the climb itself
            - length: Length of the climb in KMs.
            - steepness: Steepness of the climb in %.
            - top: Height above sea level at the top of the climb in meters.
            - km_before_finnish: KMs to finnish from the top of the climb.

        :raises ValueError: When one of args is of invalid value.
        :return: Table with wanted fields.
        """
        available_fields = (
            "climb_name",
            "climb_url",
            "length",
            "steepness",
            "top",
            "km_before_finnish"
        )
        fields = parse_table_fields_args(args, available_fields)
        table_html = self.html.css_first("table.basic")
        if table_html.css_first("tbody > tr") is None:
            return []
        table_parser = TableParser(table_html)
        casual_fields = [f for f in fields if f in ("climb_name", "climb_url")]
        table_parser.parse(casual_fields)
        if "length" in fields:
            lengths = table_parser.parse_extra_column("Length", float)
            table_parser.extend_table("length", lengths)
        if "steepness" in fields:
            lengths = table_parser.parse_extra_column("Steepness", float)
            table_parser.extend_table("steepness", lengths)
        if "top" in fields:
            lengths = table_parser.parse_extra_column("Top (m)", int)
            table_parser.extend_table("top", lengths)
        if "km_before_finnish" in fields:
            lengths = table_parser.parse_extra_column("Top at KM",
                lambda x: int(x) if x else None)
            table_parser.extend_table("km_before_finnish", lengths)
        return table_parser.table
